from typing import Any, Union, List, Optional, Dict

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util.errors import MissingDependencyError, BackendError


class GurobiBackend(Backend):
    """`Backend` implementation for the Gurobi Solver."""

    def __init__(self,
                 time_limit: Optional[float] = None,
                 solution_limit: Optional[float] = None,
                 verbose: bool = False,
                 **solver_args):
        """
        :param time_limit:
            The solver time limit.

        :param solution_limit:
            The solver solution limit.

        :param verbose:
            Whether or not to print information during the optimization process.

        :param solver_args:
            Parameters of the solver to be set via the `model.SetParam()` function.
        """
        super(GurobiBackend, self).__init__()

        try:
            import gurobipy
            self._gp = gurobipy
            """The lazily imported gurobipy instance."""
        except ModuleNotFoundError:
            raise MissingDependencyError(package='gurobipy')

        self.verbose: bool = verbose
        """Whether or not to print information during the optimization process."""

        self.solver_args: Dict[str, Any] = solver_args
        """Parameters of the solver to be set via the `model.SetParam()` function."""

        self._env: Optional = None
        """The gurobi environment instance."""

        if time_limit is not None and 'TimeLimit' not in self.solver_args:
            assert time_limit > 0, f"the time limit must be positive, got {time_limit}"
            self.solver_args['TimeLimit'] = time_limit
        if solution_limit is not None and 'SolutionLimit' not in self.solver_args:
            assert solution_limit > 0, f"the solution limit must be positive, got {solution_limit}"
            self.solver_args['SolutionLimit'] = solution_limit

    def _build_model(self) -> Any:
        self._env = self._gp.Env(empty=True)
        self._env.setParam('OutputFlag', self.verbose)
        self._env.start()
        model = self._gp.Model(env=self._env, name='model')
        for param, value in self.solver_args.items():
            model.setParam(param, value)
        return model

    def _solve_model(self) -> Optional:
        self.model.update()
        self.model.optimize()
        return None if self.model.SolCount == 0 else self.model

    def clear(self) -> Any:
        self._env.dispose()
        self.model.dispose()
        super(GurobiBackend, self).clear()

    def minimize(self, cost) -> Any:
        self.model.setObjective(cost, self._gp.GRB.MINIMIZE)
        return self

    def maximize(self, cost) -> Any:
        self.model.setObjective(cost, self._gp.GRB.MAXIMIZE)
        return self

    def get_objective(self) -> float:
        return self.solution.objVal

    def get_values(self, expressions: np.ndarray) -> np.ndarray:
        values = [v.x if isinstance(v, self._gp.Var) else v.getValue() for v in expressions.flatten()]
        return np.reshape(values, expressions.shape)

    def add_variable(self, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> Any:
        if not hasattr(self._gp.GRB, vtype.upper()):
            raise BackendError(unsupported=f"vtype '{vtype}'")
        # addVar does not accept name=None as parameter
        kwargs = dict() if name is None else dict(name=name)
        var = self.model.addVar(vtype=getattr(self._gp.GRB, vtype.upper()), lb=lb, ub=ub, **kwargs)
        return var

    def add_variables(self, *keys: int, vtype: str, lb: float, ub: float, name: Optional[str] = None) -> np.ndarray:
        if len(keys) == 0:
            # if no keys are passed, builds a single variable then reshape it into a zero-dimensional numpy array
            var = self.add_variable(vtype=vtype, lb=lb, ub=ub, name=name)
        elif not hasattr(self._gp.GRB, vtype.upper()):
            raise BackendError(unsupported=f"vtype '{vtype}'")
        else:
            vtype = getattr(self._gp.GRB, vtype.upper())
            var = self.model.addVars(*keys, vtype=vtype, lb=lb, ub=ub, name=name).values()
        return np.array(var).reshape(keys)

    def add_constraints(self, constraints: Union[List, np.ndarray], name: Optional[str] = None) -> Any:
        self.model.addConstrs((c for c in constraints), name=name)
        return self

    def add_constraint(self, constraint, name: Optional[str] = None) -> Any:
        # addConstr does not accept name=None as parameter
        kwargs = dict() if name is None else dict(name=name)
        self.model.addConstr(constraint, **kwargs)
        return self

    def add_indicator_constraint(self, indicator, expression, value: int = 1, name: Optional[str] = None) -> Any:
        # addGenConstrIndicator does not accept name=None as parameter
        kwargs = dict() if name is None else dict(name=name)
        self.model.addGenConstrIndicator(indicator, value, expression, **kwargs)

    def square(self, a, aux: Optional[str] = 'auto') -> np.ndarray:
        return self.aux(expressions=a ** 2, aux_vtype=aux)

    def sqrt(self, a, aux: Optional[str] = 'continuous') -> np.ndarray:
        a = np.atleast_1d(a)
        self._aux_warning(exp='continuous', aux=aux, msg='to compute square roots')
        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        sqrt_vector = self.add_continuous_variables(*a.shape, lb=0.0, ub=float('inf'))
        for aux_var, sqrt_var in zip(aux_vector.flatten(), sqrt_vector.flatten()):
            self.model.addGenConstrPow(sqrt_var, aux_var, 2)
        return sqrt_vector

    def abs(self, a: np.ndarray, aux: Optional[str] = 'continuous') -> np.ndarray:
        a = np.atleast_1d(a)
        self._aux_warning(exp='continuous', aux=aux, msg='to compute absolute values')
        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        abs_vector = self.add_continuous_variables(*a.shape, lb=0.0, ub=float('inf'))
        for aux_var, abs_var in zip(aux_vector.flatten(), abs_vector.flatten()):
            self.model.addGenConstrAbs(abs_var, aux_var)
        return abs_vector

    def log(self, a: np.ndarray, aux: Optional[str] = 'continuous') -> np.ndarray:
        a = np.atleast_1d(a)
        self._aux_warning(exp='continuous', aux=aux, msg='to compute logarithms')
        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        log_vector = self.add_continuous_variables(*a.shape, lb=-float('inf'), ub=float('inf'))
        for aux_var, log_var in zip(aux_vector.flatten(), log_vector.flatten()):
            self.model.addGenConstrExp(log_var, aux_var)
        return log_vector

    def min(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = 'auto') -> Any:
        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        vtype = 'continuous' if aux == 'auto' else aux
        lb, ub = (0, 1) if aux == 'binary' else (-float('inf'), float('inf'))

        def _min(_array):
            min_val = self.add_variable(vtype=vtype, lb=lb, ub=ub)
            self.model.addGenConstrMin(min_val, list(_array), constant=float('inf'))
            return min_val

        return self._handle_axes(aux_vector, operation=_min, axis=axis, asarray=asarray)

    def max(self, a: np.ndarray, axis: Optional[int] = None, asarray: bool = False, aux: Optional[str] = 'auto') -> Any:
        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        vtype = 'continuous' if aux == 'auto' else aux
        lb, ub = (0, 1) if aux == 'binary' else (-float('inf'), float('inf'))

        def _max(_array):
            max_val = self.add_variable(vtype=vtype, lb=lb, ub=ub)
            self.model.addGenConstrMax(max_val, list(_array), constant=-float('inf'))
            return max_val

        return self._handle_axes(aux_vector, operation=_max, axis=axis, asarray=asarray)

    def divide(self, a: np.ndarray, b: np.ndarray, aux: Optional[str] = 'auto'):
        try:
            return super(GurobiBackend, self).divide(a, b, aux=aux)
        except self._gp.GurobiError:
            # gurobipy.GurobiError: Divisor must be a constant
            # in case the divisor is not an array of constants, we handle the case by adding N auxiliary variables z_i
            # so that a_i / b_i = z_i --> a_i = z_i * b_i
            self._aux_warning(exp='continuous', aux=aux, msg='to compute divisions with non-constant divisors')
            a, b = np.atleast_1d(a), np.atleast_2d(b)
            z = self.add_continuous_variables(*a.shape)
            self.add_constraints([ai == zi * bi for ai, bi, zi in zip(a.flatten(), b.flatten(), z.flatten())])
            return z

    def norm_0(self,
               a: np.ndarray,
               axis: Optional[int] = None,
               asarray: bool = False,
               aux: Optional[str] = 'auto') -> Any:
        def _norm(_array):
            norm_val = self.add_continuous_variable(lb=-float('inf'), ub=float('inf'))
            self.model.addGenConstrNorm(norm_val, list(_array), which=0)
            return norm_val

        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        self._aux_warning(exp='continuous', aux=aux, msg='to compute norm 1')
        return self._handle_axes(aux_vector, operation=_norm, axis=axis, asarray=asarray)

    def norm_1(self,
               a: np.ndarray,
               axis: Optional[int] = None,
               asarray: bool = False,
               aux: Optional[str] = 'auto') -> Any:
        def _norm(_array):
            norm_val = self.add_continuous_variable(lb=-float('inf'), ub=float('inf'))
            self.model.addGenConstrNorm(norm_val, list(_array), which=1)
            return norm_val

        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        self._aux_warning(exp='continuous', aux=aux, msg='to compute norm 1')
        return self._handle_axes(aux_vector, operation=_norm, axis=axis, asarray=asarray)

    def norm_2(self,
               a,
               squared: bool = False,
               axis: Optional[int] = None,
               asarray: bool = False,
               aux: Optional[str] = 'auto') -> Any:
        if squared:
            def _norm(_array):
                norm_val = self.add_continuous_variable(lb=-float('inf'), ub=float('inf'))
                self.model.addGenConstrNorm(norm_val, list(_array), which=2)
                return self.square(norm_val)
        else:
            def _norm(_array):
                norm_val = self.add_continuous_variable(lb=-float('inf'), ub=float('inf'))
                self.model.addGenConstrNorm(norm_val, list(_array), which=2)
                return norm_val

        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        self._aux_warning(exp='continuous', aux=aux, msg='to compute norm 2')
        return self._handle_axes(aux_vector, operation=_norm, axis=axis, asarray=asarray)

    def norm_inf(self,
                 a: np.ndarray,
                 axis: Optional[int] = None,
                 asarray: bool = False,
                 aux: Optional[str] = 'auto') -> Any:
        def _norm(_array):
            from gurobipy import GRB
            norm_val = self.add_continuous_variable(lb=-float('inf'), ub=float('inf'))
            self.model.addGenConstrNorm(norm_val, list(_array), which=GRB.INFINITY)
            return norm_val

        # creating auxiliary variables is necessary since 'addGenConstr' does not accept expressions
        aux_vector = self.aux(expressions=a, aux_vtype='continuous')
        self._aux_warning(exp='continuous', aux=aux, msg='to compute norm inf')
        return self._handle_axes(aux_vector, operation=_norm, axis=axis, asarray=asarray)



"""Learners to fit custom curves."""
from typing import Optional, Union, List, Callable

import numpy as np
import scipy.optimize

from moving_targets.learners.learner import Learner
from moving_targets.util.scalers import Scaler


class CurveLearner(Learner):
    """Template class for a learner which fits a custom curve."""

    def __init__(self,
                 curve: Callable,
                 method: Callable,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False):
        """
        :param curve:
            A function f(x; p) -> y which is fed with an input vector (x) and the learnable parameters (p).

        :param method:
            An optimization routine m(f; x, y) -> p* which is fed with the curve function (f) and the training data
            (x, y), and eventually returns the optimal configuration for the parameters (p*).

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.
        """
        super(CurveLearner, self).__init__(mask=mask, x_scaler=x_scaler, y_scaler=y_scaler, stats=stats)

        self.curve: Callable = curve
        """The function to be learned."""

        self.method: Callable = method
        """The optimization routine."""

        self.configuration: Optional = None
        """The optimal configuration of the curve parameters."""

    def _fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None):
        if sample_weight is not None:
            self.logger.warning("CurveLearner does not support sample weights, please pass 'sample_weight'=None")
        self.configuration = self.method(self.curve, x, y)

    def _predict(self, x) -> np.ndarray:
        return self.curve(x, *self.configuration)


class ScipyCurveFit(CurveLearner):
    def __init__(self,
                 curve: Callable,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **method_kwargs):
        """
        :param curve:
            A function f(x; p) -> y which is fed with an input vector (x) and the learnable parameters (p).

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param method_kwargs:
            Additional arguments to be passed to `scipy.optimize.curve_fit`.
        """

        def method(f, x, y):
            return scipy.optimize.curve_fit(f=f, xdata=x, ydata=y, **method_kwargs)[0]

        super(ScipyCurveFit, self).__init__(curve=curve,
                                            method=method,
                                            mask=mask,
                                            x_scaler=x_scaler,
                                            y_scaler=y_scaler,
                                            stats=stats)



"""Moving Targets Losses."""
from typing import Optional, Any, Callable, Tuple

import numpy as np

from moving_targets.masters.backends import Backend
from moving_targets.util import probabilities
from moving_targets.util.errors import not_implemented_message
from moving_targets.util.masking import mask_data, get_mask


class Loss:
    """Basic interface for a Moving Targets Master Loss."""

    def __init__(self, binary: bool, name: str):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param name:
            The name of the loss.
        """
        # handle squared values computation based on whether or not variables are binary, indeed, given the model
        # variables <v> and the learners predictions <p>, for binary variables we have that:
        #    square(v, p) = [v * (1 - p) + (1 - v) * p]^2 =
        #                 = [v * (1 - p)]^2        + [(1 - v) * p]^2	    - 2 * [v * (1 - p) * (1 - v) * p] =
        #                 = [v^2 * (1 - 2p + p^2)] + [p^2 * (1 - 2v + v^2)] - 2 * [v * (1 - v) * (1 - p) * p] =
        #                 = [v - 2vp + vp^2]       + [p^2 - 2vp^2 + vp^2]   - 2 * [(v - v) * (p - p^2)] = --> as v^2 = v
        #                 = [v - 2vp + vp^2]       + [p^2 - vp^2]           - 2 * [0] =
        #                 =  v - 2vp + p^2
        # thus we can reformulate this as a linear expressions, which can be handled in a faster way
        if binary:
            square = lambda b, v, p: v - 2 * v * p + p ** 2
        else:
            square = lambda b, v, p: b.square(v - p)

        self.binary: bool = binary
        """Whether the model variables are expected to be binary or not."""

        self.square: Callable = square
        """A callable function of type f(<b>, <v>, <p>) -> <(p - v) ** 2> representing the square strategy, where <b>
        is the backend instance, <v> is the array of backend variables, and <p> is the array of predictions."""

        self.__name__: str = name
        """The name of the loss."""

    def __call__(self,
                 backend: Backend,
                 variables: np.ndarray,
                 targets: np.ndarray,
                 predictions: np.ndarray,
                 sample_weight: Optional[np.ndarray] = None,
                 mask: Optional[float] = None) -> Tuple[Any, Any]:
        """Computes the two terms of the loss value, which are obtained by approximating the actual loss <L> to its
        first order taylor expansion L(z, y) in the point <p>, where <z> are the model variables, <y> the original
        targets, and <p> the learners predictions.

        The taylor expansion of L(z, y) at point p is defined as: L(p, y) + dL(p, y) / dp * (z - p) + o((z - p) ^ 2).
        Since the term "L(p, y)" is constant, we can discard it when minimizing <L>, thus the final loss can be seen as
        a balance between the term "dL(p, y) / dp * (z - p)" and the term "(z - p) ^ 2", which is achieved through the
        balancing parameter <alpha>, thus obtaining: min_z {alpha * dL(p, y) / dp * (z - p) + (z - p) ^ 2}.

        In case of multiple features, the term dL(p, y) / dp * (z - p) is replaced by the sum of partial derivatives
        dL(p, y) / dp_i multiplied by the respective component (z_i - p_i), while the (z - p) ^ 2 term will be computed
        as a mean over all the features according to the norm-two definition, leading to a final formulation which is:
        min_z {alpha * (dL(p, y) / dp_1 * (z_1 - p_1) + ... + dL(p, y) / dp_F * (z_F - p_F)) + |z - p|_2^2}, or else:
        min_z {alpha * sum(nabla_L(p, y) * (z - p)) + (z - p) ^ 2}.

        :param backend:
            The `Backend` instance used to compute the loss.

        :param variables:
            The array of backend model variables (z).

        :param targets:
            The array of original targets (y).

        :param predictions:
            The array of learners predictions (p).

        :param sample_weight:
            The (optional) array of sample weights.

        :param mask:
            An (optional) masking value used to mask the original targets.

        :return:
            A tuple containing the nabla term nabla_L(p, y) @ (z - p)^T and the squared term (z - p) @ (z - p)^T.
        """
        mask = get_mask(targets, mask)
        # reshape inputs in order to create a NxF matrix, where N is the number of samples and F the number of features
        variables = np.reshape(variables, (len(variables), -1))
        targets = np.reshape(targets, (len(targets), -1))
        predictions = np.reshape(predictions, (len(predictions), -1))
        # flatten sample weights (or create a vector of ones is none are passed), then normalize them
        sample_weight = np.ones(len(variables)) if sample_weight is None else sample_weight.flatten()
        sample_weight = len(sample_weight) * np.array(sample_weight) / np.sum(sample_weight)
        # compute the nabla term, which is a summation of the terms dL / dp_i * (z_i - p_i), with i in {1, ..., F}
        # also mask each vector using the previously computed mask as reference
        m_targ, m_pred, m_var, m_weights = mask_data(targets, predictions, variables, sample_weight, mask=mask)
        nabla_term = self.nabla(targets=m_targ, predictions=m_pred)
        nabla_term = backend.sum(nabla_term * (m_var - m_pred), axis=1)
        nabla_term = backend.mean(m_weights * nabla_term)
        # compute the squared term, which is the mean of the terms (z_i - p_i) ^ 2, with i in {1, ..., F}
        squared_term = self.square(b=backend, v=variables, p=predictions)
        squared_term = backend.mean(squared_term, axis=1)
        squared_term = backend.mean(sample_weight * squared_term)
        # finally return the two losses
        return nabla_term, squared_term

    def nabla(self, targets: np.ndarray, predictions: np.ndarray) -> np.ndarray:
        """Computes the nabla_L(p, y) values, with <p> being the predictions and <y> being the original targets.

        :param targets:
            The array of original targets (y).

        :param predictions:
            The array of learners predictions (p).

        :return:
            The array of values of nabla_L(p, y) for each of the samples.
        """
        raise NotImplementedError(not_implemented_message(name='nabla'))


class MAE(Loss):
    """Mean Absolute Error Loss."""

    def __init__(self, binary: bool = False, name: str = 'mean_absolute_error'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

        :param name:
            The name of the loss.
        """
        super(MAE, self).__init__(binary=binary, name=name)

    def nabla(self, targets: np.ndarray, predictions: np.ndarray) -> np.ndarray:
        # given a target vector with <F> features, the loss is: L(p, t) = sum(|p - t|) / F
        # thus, its partial derivative will be: dL / dp_i = sign(p - t) / F
        return np.sign(predictions - targets) / targets.shape[1]


class MSE(Loss):
    """Mean Squared Error Loss."""

    def __init__(self, binary: bool = False, name: str = 'mean_squared_error'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

            Knowing that the model variables are binary allows to deal with error computation in a faster way since it
            will lead to a simplified scenario in which the model variables can only have an integer value in {0, 1},
            while the numeric variables can only have a continuous value in [0, 1].

        :param name:
            The name of the loss.
        """
        super(MSE, self).__init__(binary=binary, name=name)

    def nabla(self, targets: np.ndarray, predictions: np.ndarray) -> np.ndarray:
        # given a target vector with <F> features, the loss is: L(p, t) = sum((p - t) ^ 2) / F
        # thus, its partial derivative will be: dL / dp_i = 2 * (p - t) / F
        return 2 * (predictions - targets) / targets.shape[1]


class HammingDistance(Loss):
    """Hamming Distance."""

    def __init__(self, labelling: bool = False, name: str = 'hamming_distance'):
        """
        :param labelling:
            Whether this is a loss for a labelling or a classification task.

            This parameter is used to retrieve classes/labels from the vector of predicted probabilities in multiclass
            or multilabel scenarios. For binary tasks this has no effect.

        :param name:
            Whether the master represents a labelling or a classification task.
        """
        super(HammingDistance, self).__init__(binary=True, name=name)

        self.labelling: bool = labelling
        """Whether this is a loss for a labelling or a classification task."""

    def nabla(self, targets: np.ndarray, predictions: np.ndarray) -> np.ndarray:
        # given a target vector with <F> features, the loss is: L(p, t) = sum((1 - t) * p + t * (1 - p)) / F
        # which can be eventually rewritten as: L(p, t) = sum(p - 2 * t * p + t) / F
        # thus, its partial derivative will be: dL / dp_i = (1 - 2 * t_i) / F
        return (1 - 2 * targets) / targets.shape[1]

    def __call__(self,
                 backend: Backend,
                 variables: np.ndarray,
                 targets: np.ndarray,
                 predictions: np.ndarray,
                 sample_weight: Optional[np.ndarray] = None,
                 mask: Optional[float] = None) -> Tuple[Any, Any]:
        assert np.all(targets == targets.astype(int)), f"HammingDistance can handle integer targets only, got {targets}"
        # since the hamming distance works on discrete targets only, we retrieve the class values from the predictions
        if self.labelling:
            # in case of labelling tasks, this is enough to get a bi-dimensional matrix which matches the targets shape
            predictions = probabilities.get_classes(prob=predictions, labelling=True)
        else:
            # otherwise, if we have a multiclass task we need to onehot encode the probabilities to match the shape
            classes = 2 if targets.squeeze().ndim == 1 else targets.shape[1]
            predictions = probabilities.get_classes(prob=predictions, labelling=False)
            predictions = probabilities.get_onehot(vector=predictions, classes=classes)
        return super(HammingDistance, self).__call__(backend, variables, targets, predictions, sample_weight, mask)


class CrossEntropy(Loss):
    """Negative Log-Likelihood Loss."""

    def __init__(self, binary: bool = True, clip_value: float = 1e-3, name: str = 'crossentropy'):
        """
        :param binary:
            Whether the model variables are expected to be binary or not.

        :param clip_value:
            The clipping value to be used to avoid numerical errors.

        :param name:
            The name of the metric.
        """
        super(CrossEntropy, self).__init__(binary=binary, name=name)

        self.clip_value: float = clip_value
        """The clipping value to be used to avoid numerical errors."""

    def nabla(self, targets: np.ndarray, predictions: np.ndarray) -> np.ndarray:
        # given a target vector with <F> features, the loss is: L(p, t) = sum(-t * ln(p))
        # thus, its partial derivative will be: dL / dp_i = -t / p
        return -targets / predictions.clip(min=self.clip_value, max=1 - self.clip_value)

    def __call__(self,
                 backend: Backend,
                 variables: np.ndarray,
                 targets: np.ndarray,
                 predictions: np.ndarray,
                 sample_weight: Optional[np.ndarray] = None,
                 mask: Optional[float] = None) -> Tuple[Any, Any]:
        # in order to deal with both binary and categorical crossentropy, if the prediction array has either size (N,)
        # or (N, 1), we convert it (along with the targets and variables vectors) to the form [1 - a, a]
        if predictions.squeeze().ndim == 1:
            variables = np.concatenate([1 - variables.reshape((-1, 1)), variables.reshape((-1, 1))], axis=1)
            targets = np.concatenate([1 - targets.reshape((-1, 1)), targets.reshape((-1, 1))], axis=1)
            predictions = np.concatenate([1 - predictions.reshape((-1, 1)), predictions.reshape((-1, 1))], axis=1)
        return super(CrossEntropy, self).__call__(backend, variables, targets, predictions, sample_weight, mask)


aliases: dict = {
    # Mean Absolute Error
    'mae': MAE,
    'mean_absolute_error': MAE,
    'mean absolute error': MAE,
    # Mean Squared Error
    'mse': MSE,
    'mean_squared_error': MSE,
    'mean squared error': MSE,
    # Hamming Distance
    'hd': HammingDistance,
    'hamming_distance': HammingDistance,
    'hamming distance': HammingDistance,
    'bhd': HammingDistance,
    'binary_hamming': HammingDistance,
    'binary hamming': HammingDistance,
    'chd': HammingDistance,
    'categorical_hamming': HammingDistance,
    'categorical hamming': HammingDistance,
    # CrossEntropy
    'ce': CrossEntropy,
    'crossentropy': CrossEntropy,
    'bce': CrossEntropy,
    'binary_crossentropy': CrossEntropy,
    'binary crossentropy': CrossEntropy,
    'cce': CrossEntropy,
    'categorical_crossentropy': CrossEntropy,
    'categorical crossentropy': CrossEntropy,
    'll': CrossEntropy,
    'log_likelihood': CrossEntropy,
    'log likelihood': CrossEntropy,
    'nll': CrossEntropy,
    'negative_log_likelihood': CrossEntropy,
    'negative log likelihood': CrossEntropy,
}
"""Dictionary which associates to each loss alias the respective class type."""



"""Utility functions to handle classes and probabilities"""

from typing import Optional

import numpy as np


def count_classes(vector: np.ndarray) -> int:
    """Counts the number of classes in a vector of class targets in the range {0, ..., C - 1}.

    :param vector:
        A vector of class values.

    :return:
        The number of classes.
    """
    return int(np.max(vector)) + 1


def get_classes(prob: np.ndarray, labelling: bool = False) -> np.ndarray:
    """Gets the output classes given the output probabilities per class.

    :param prob:
        An array of probabilities.

    :param labelling:
        Whether the task that generated these probabilities is a labelling or a classification task. This defines the
        strategy to get the discrete values from probabilities in case the given array is bi-dimensional (indeed, for
        one-dimensional vectors this parameter is ignored since the two tasks will be the same), which consists in
        rounding or taking the the argmax over each row, respectively.

    :return:
        The respective output classes/labels.
    """
    return prob.round().astype(int) if prob.squeeze().ndim == 1 or labelling else prob.argmax(axis=1)


def get_onehot(vector: np.ndarray, classes: Optional[int] = None, onehot_binary: bool = False) -> np.ndarray:
    """Gets the one hot encoded representation of the given classes.

    :param vector:
        A vector of class values.

    :param classes:
        Either an integer representing the number of categories or None, in which case the labels are inferred.

    :param onehot_binary:
        Whether or not to onehot encode a vector of binary variables.

    :return:
        The vector (in case the number of classes is 2) or matrix of binary variables.
    """
    vector = vector.squeeze()
    assert vector.ndim == 1, "Pleas provide a be one-dimensional vector"
    assert np.issubdtype(vector.dtype, np.integer), "Please provide a vector of integer numbers only"
    assert np.min(vector) >= 0, "Please provide a vector of non-negative class identifiers"
    # retrieves the correct number of samples and classes depending on the input
    samples = len(vector)
    classes = classes or count_classes(vector=vector)
    # if the task is binary there is nothing to encode, otherwise creates a SxC matrix and places the ones accordingly.
    if classes == 2 and not onehot_binary:
        return vector
    else:
        onehot = np.zeros((samples, classes), dtype=int)
        onehot[np.arange(samples), vector] = 1
        return onehot



"""Type aliases."""

from typing import Tuple, Dict, Any

import numpy as np

Dataset = Dict[str, Tuple[Any, np.ndarray]]
"""A dictionary where the key is string representing the name of the split, and the value is tuple (x, y)."""


def is_numeric(obj: Any) -> bool:
    """Checks whether an object has a numeric type or not.

    :param obj:
        The object to test.

    :return:
        True if the object has a numeric type, False otherwise.
    """
    return isinstance(obj, (int, float, np.number))



"""Basic Master Interface."""
import time
from abc import ABC
from typing import Set, Optional
from typing import Union, List

import numpy as np

from moving_targets.callbacks import StatsLogger
from moving_targets.masters import optimizers, losses, backends
from moving_targets.masters.backends import Backend
from moving_targets.masters.losses import MAE, MSE, Loss, HammingDistance
from moving_targets.masters.optimizers import Constant, Optimizer
from moving_targets.util import probabilities
from moving_targets.util.errors import not_implemented_message
from moving_targets.util.masking import get_mask
from moving_targets.util.scalers import Scaler
from moving_targets.util.typing import Dataset


class Master(StatsLogger):
    """Basic interface for a Moving Targets Master."""

    @staticmethod
    def _parameters() -> Set[str]:
        return {'alpha', 'nabla_term', 'squared_term', 'objective', 'elapsed_time'}

    def __init__(self,
                 backend: Union[str, Backend],
                 loss: Union[str, Loss],
                 alpha: Union[str, float, Optimizer],
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param loss:
            Either a string representing a `Loss` alias or the actual `Loss` instance used to compute the objective.

        :param alpha:
            Either a floating point for a constant alpha value, a string representing an `Optimizer` alias,  or an
            actual `Optimizer` instance which implements the strategy to dynamically change the alpha value.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
           'nabla_term', 'squared_term', 'objective', 'elapsed_time'] whose statistics must be logged.
        """

        super(Master, self).__init__(stats=stats, name='Master')

        # handle alpha optimizer
        if isinstance(alpha, str):
            alpha = self._get_optimizer(optimizer=alpha)
        elif not isinstance(alpha, Optimizer):
            alpha = Constant(base=alpha)

        self.backend: Backend = backend if isinstance(backend, Backend) else self._get_backend(backend=backend)
        """The `Backend` instance."""

        self.loss: Loss = loss if isinstance(loss, Loss) else self._get_loss(loss=loss)
        """The `Loss` instance."""

        self.alpha: Optimizer = alpha
        """The alpha `Optimizer` instance."""

        self.mask: Optional[float] = mask
        """The (optional) masking value used to mask the original targets."""

        self.x_scaler: Optional[Scaler] = Scaler(default_method=x_scaler) if isinstance(x_scaler, str) else x_scaler
        """The (optional) scaler for the input data."""

        self.y_scaler: Optional[Scaler] = Scaler(default_method=y_scaler) if isinstance(y_scaler, str) else y_scaler
        """The (optional) scaler for the output data."""

        self._macs: Optional = None
        """Reference to the MACS object encapsulating the `Master`."""

        self._time: Optional[float] = None
        """An auxiliary variable to keep track of the elapsed time between iterations."""

    # noinspection PyMethodMayBeStatic
    def _get_backend(self, backend: str) -> Backend:
        """Implements the strategy to get default backends from the class type retrieved using the backend alias.

        :param backend:
            The backend alias.

        :return:
            The correct `Optimizer` instance.
        """
        backend_class = backends.aliases.get(backend)
        assert backend_class is not None, f"Unknown backend alias '{backend}'"
        return backend_class()

    def _get_loss(self, loss: str) -> Loss:
        """Implements the strategy to get default losses from the class type retrieved using the loss alias.

        :param loss:
            The loss alias.

        :return:
            The correct `Loss` instance.
        """
        loss_class = losses.aliases.get(loss)
        assert loss_class is not None, f"Unknown loss alias '{loss}'"
        return loss_class(name=loss)

    # noinspection PyMethodMayBeStatic
    def _get_optimizer(self, optimizer: str) -> Optimizer:
        """Implements the strategy to get default optimizers from the class type retrieved using the optimizer alias.

        :param optimizer:
            The optimizer alias.

        :return:
            The correct `Optimizer` instance.
        """
        optimizer_class = optimizers.aliases.get(optimizer)
        assert optimizer_class is not None, f"Unknown alpha optimizer alias '{optimizer}'"
        return optimizer_class(base=1)

    def log(self, **cache):
        self._macs.log(**cache)

    def on_adjustment_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._macs = macs
        self._time = time.time()

    def on_adjustment_end(self, macs, x, y: np.ndarray, z: np.ndarray, val_data: Optional[Dataset]):
        self._log_stats(elapsed_time=time.time() - self._time)
        self._time = None
        self._macs = None

    def on_backend_solved(self, x, y: np.ndarray, p: np.ndarray, z: np.ndarray):
        """Routine called after the master problem is solved and the adjusted targets are retrieved.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param p:
            The vector of learners predictions.

        :param z:
            The vector of adjusted labels.
        """
        pass

    def build(self, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        """Creates the model variables and adds the problem constraints.

        :param x:
            The matrix/dataframe of data samples.

        :param y:
            The vector of original labels.

        :param p:
            The vector of learner predictions.

        :return:
            The model variables.
        """
        raise NotImplementedError(not_implemented_message(name='build'))

    def adjust_targets(self,
                       x,
                       y: np.ndarray,
                       p: Optional[np.ndarray],
                       sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
        """Core function of the `Master` object which builds the model and returns the adjusted targets.

        :param x:
            The matrix/dataframe of data samples.

        :param y:
            The vector of original labels.

        :param p:
            The (optional) vector of learner predictions, which will be None in the first iteration with 'projection'.

        :param sample_weight:
            The (optional) vector of sample weights.

        :return:
            The vector of adjusted targets.
        """
        # use the original targets when no predictions are available due to the initial projection step
        p = y if p is None else p
        # transform data if scalers are passed
        x = x if self.x_scaler is None else self.x_scaler.fit_transform(data=x)
        if self.y_scaler is not None:
            # this is a sort of workaround: the masking of the vectors is delegated to the loss instance, thus there
            # should be no need to mask the vector at this point, however, since here we want to scale the target data
            # which might be masked later, we fit the scaler only on the masked data (this will avoid numerical errors
            # in case of nan masking, or wrong scaling factors in case of value-based or vector-based masking)
            self.y_scaler.fit(y[get_mask(y, self.mask)])
            y = self.y_scaler.transform(data=y)
            p = self.y_scaler.transform(data=p)
        # build backend and model
        self.backend.build()
        v = self.build(x=x, y=y, p=p)
        alpha = self.alpha(x=x, y=y, p=p)
        nabla_term, squared_term = self.loss(backend=self.backend,
                                             variables=v,
                                             targets=y,
                                             predictions=p,
                                             sample_weight=sample_weight,
                                             mask=self.mask)
        self.backend.minimize(cost=alpha * nabla_term + squared_term)
        # if the problem is infeasible return None, otherwise log stats and return the adjusted labels
        if self.backend.solve().solution is not None:
            self._log_stats(alpha=alpha,
                            nabla_term=self.backend.get_value(nabla_term),
                            squared_term=self.backend.get_value(squared_term),
                            objective=self.backend.get_objective())
            adjusted = self.backend.get_values(expressions=v)
        else:
            self._log_stats(alpha=alpha)
            adjusted = None
        self.on_backend_solved(x=x, y=y, p=p, z=adjusted)
        self.backend.clear()
        # invert transformation on output targets if scaler is passed
        return adjusted if self.y_scaler is None else self.y_scaler.inverse_transform(adjusted)


class RegressionMaster(Master, ABC):
    """An abstract Moving Targets Master for Regression Tasks."""

    def __init__(self, backend: Union[str, Backend],
                 loss: Union[str, Loss] = 'mse',
                 alpha: Union[str, float, Optimizer] = 'harmonic',
                 lb: float = -float('inf'),
                 ub: float = float('inf'),
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param loss:
            Either a string representing a `Loss` alias or the actual `Loss` instance used to compute the objective.

        :param alpha:
            Either a floating point for a constant alpha value, a string representing an `Optimizer` alias,  or an
            actual `Optimizer` instance which implements the strategy to dynamically change the alpha value.

        :param lb:
            The model variables lower bounds.

        :param ub:
            The model variables upper bounds.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
            'nabla_term', 'squared_term', 'objective', 'elapsed_time'] whose statistics must be logged.
        """
        super(RegressionMaster, self).__init__(backend=backend,
                                               loss=loss,
                                               alpha=alpha,
                                               mask=mask,
                                               x_scaler=x_scaler,
                                               y_scaler=y_scaler,
                                               stats=stats)

        self.lb = lb
        """The model variables lower bounds."""

        self.ub = ub
        """The model variables upper bounds."""

    def _get_loss(self, loss: str) -> Loss:
        loss_class = losses.aliases.get(loss)
        assert loss_class is not None, f"Unknown loss alias '{loss}'"
        assert loss_class in [MAE, MSE], f"Loss '{loss} is not a valid regression loss."
        return loss_class(name=loss)

    def build(self, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        return self.backend.add_continuous_variables(*y.shape, lb=self.lb, ub=self.ub, name='y')


class ClassificationMaster(Master, ABC):
    """An abstract Moving Targets Master for Classification Tasks."""

    def __init__(self,
                 backend: Union[str, Backend],
                 loss: Union[str, Loss] = 'crossentropy',
                 alpha: Union[str, float, Optimizer] = 'harmonic',
                 labelling: bool = False,
                 types: str = 'auto',
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False):
        """
        :param backend:
            The `Backend` instance encapsulating the optimization solver.

        :param loss:
            Either a string representing a `Loss` alias or the actual `Loss` instance used to compute the objective.

        :param alpha:
            Either a floating point for a constant alpha value, a string representing an `Optimizer` alias,  or an
            actual `Optimizer` instance which implements the strategy to dynamically change the alpha value.

        :param labelling:
            Whether this is a labelling or a classification task.

        :param types:
            The variables and adjustments types, which must be in ['auto', 'discrete', 'discretized', 'continuous'].

            - if 'discrete' is chosen, the model will use binary variables to represent class targets, and the returned
            adjustments will be discrete (i.e., a matrix of binary labels in case of labelling tasks, or a vector of
            categorical values in case of classification tasks);
            - if 'discretized' is chosen, the model will use continuous variables to represent class probabilities, but
            the returned adjustments will be discretized in order to match the same return type of 'discrete';
            - if 'continuous' is chosen, the model will use continuous variables to represent class probabilities, and
            the returned adjustments will be continuous as well (i.e., a matrix of binary class/label probabilities for
            both multiclass and multilabel tasks, or a vector of probabilities for binary tasks);
            - if 'auto' is chosen, the model will go for the 'discrete' option unless an explicit loss instance is
            passed, since in that case it will leverage the 'binary' field of the loss to choose the variables types
            while returning discrete adjustments.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in ['alpha',
           'nabla_term', 'squared_term', 'objective', 'elapsed_time'] whose statistics must be logged.
        """
        if types == 'auto':
            binary, classes = loss.binary if isinstance(loss, Loss) else True, True
        elif types == 'discrete':
            binary, classes = True, True
        elif types == 'discretized':
            binary, classes = False, True
        elif types == 'continuous':
            binary, classes = False, False
        else:
            raise AssertionError("'vtype' should be one in 'discrete', 'discretized', or 'continuous', got '{vtype}'")

        self.binary: bool = binary
        """Whether to use binary or continuous model variables."""

        self.classes: bool = classes
        """Whether to return class targets or class probabilities."""

        self.labelling: bool = labelling
        """Whether this is a labelling or a classification task."""

        super(ClassificationMaster, self).__init__(backend=backend,
                                                   loss=loss,
                                                   alpha=alpha,
                                                   mask=mask,
                                                   x_scaler=x_scaler,
                                                   y_scaler=y_scaler,
                                                   stats=stats)

    def _get_loss(self, loss: str) -> Loss:
        loss_class = losses.aliases.get(loss)
        assert loss_class is not None, f"Unknown loss alias '{loss}'"
        # by default, use continuous values for each loss but hamming distance that cannot handle them
        if loss_class == HammingDistance:
            assert self.binary, f"{loss} loss can handle only discrete targets, please use types = 'discrete'"
            return HammingDistance(labelling=self.labelling, name=loss)
        else:
            return loss_class(binary=self.binary, name=loss)

    def build(self, x, y: np.ndarray, p: np.ndarray) -> np.ndarray:
        # we can simply build variables by shape (the y targets will be already onehot encoded for multiclass tasks)
        vtype = 'binary' if self.binary else 'continuous'
        variables = self.backend.add_variables(*y.shape, vtype=vtype, lb=0, ub=1, name='y')
        if variables.ndim == 2 and not self.labelling:
            # if we are dealing with multiclass classification (i.e., we have a bi-dimensional array of variables but
            # we are not in the multilabel scenario), we constraint variables in each row to sum up to one in order to
            # correctly represent class probabilities
            self.backend.add_constraints([self.backend.sum(row) == 1 for row in variables])
        return variables

    def adjust_targets(self,
                       x,
                       y: np.ndarray,
                       p: Optional[np.ndarray],
                       sample_weight: Optional[np.ndarray] = None) -> np.ndarray:
        # when dealing with multilabel classification we expect the target to match the predictions shape since they
        # should both have shape (N, C), otherwise, in standard classification cases the original targets are expected
        # to be a N-sized vector while the predictions may be either a N-size vector (in case of binary classification)
        # or a NxC matrix (in case of multiclass classification), therefore in this case we onehot encode the original
        # targets to match the shape, since onehot encoding a binary target vector will make no difference
        y = y if self.labelling else probabilities.get_onehot(y)
        # due to numerical tolerances, targets may be returned as <z + eps>, therefore we round binary targets in order
        # to remove these numerical errors and make sure that they will not cause problems to the learners
        z = super(ClassificationMaster, self).adjust_targets(x, y, p, sample_weight)
        z = z.round() if self.binary else z
        # eventually, we return class/label targets if we are asked so, or class/label probabilities otherwise
        return probabilities.get_classes(z, labelling=self.labelling) if self.classes else z



"""Console Logger Callback."""

import time
from typing import Optional

import numpy as np

from moving_targets.callbacks.callback import Callback
from moving_targets.util.typing import Dataset


class ConsoleLogger(Callback):
    """Callback which logs basic information on screen during the `MACS` training."""

    def __init__(self):
        """"""
        super(ConsoleLogger, self).__init__()

        self._time: Optional[float] = None
        """An internal variable used to compute elapsed time between routines."""

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        print('Starting the MACS process...')

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # Prints the current iteration and stores the initial time.
        print(f'  > it. {macs.iteration:02}:', end=' ')
        self._time = time.time()

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # Prints the time elapsed from the iteration start.
        print(f'{time.time() - self._time:.4f} s')
        self._time = None

    def on_process_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        print('... process ended correctly!')



"""Predefined metrics for classification, regression, and constraint satisfaction for Moving Targets."""

from moving_targets.metrics.classification import Accuracy, AUC, ClassificationMetric, CrossEntropy, F1, Precision
from moving_targets.metrics.classification import Recall
from moving_targets.metrics.constraints import ClassFrequenciesStd, DIDI, MonotonicViolation, CausalIndependence
from moving_targets.metrics.metric import Metric
from moving_targets.metrics.regression import MAE, MSE, R2, RegressionMetric



"""Sklearn-based Learners."""
from typing import Optional, Union, List, Any

import numpy as np
import sklearn.ensemble as ens
import sklearn.linear_model as lm
from sklearn.preprocessing import PolynomialFeatures

from moving_targets.learners.learner import Learner
from moving_targets.util.scalers import Scaler


class ScikitLearner(Learner):
    """Wrapper for a custom Scikit-Learn model."""

    def __init__(self,
                 model,
                 polynomial: Union[None, int, PolynomialFeatures] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **fit_kwargs):
        """
        :param model:
            The Scikit-Learn model.

        :param polynomial:
            The polynomial feature preprocessor. It can be either None for no preprocessing, an actual scikit learn
            `PolynomialFeatures` instance, or an integer representing the degree used to generate polynomial features
             which will be used to create a `PolynomialFeatures` object with no bias and all interactions.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param fit_kwargs:
            Custom arguments to be passed to the model '.fit()' method.
        """
        super(ScikitLearner, self).__init__(mask=mask, x_scaler=x_scaler, y_scaler=y_scaler, stats=stats)

        if isinstance(polynomial, int):
            polynomial = PolynomialFeatures(degree=polynomial, interaction_only=False, include_bias=False)

        self.model = model
        """The Scikit-Learn model."""

        self.polynomial: Optional[PolynomialFeatures] = polynomial
        """The (optional) polynomial features preprocessor."""

        self.fit_kwargs = fit_kwargs
        """Custom arguments to be passed to the model '.fit()' method."""

    def fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None) -> Any:
        x = x if self.polynomial is None else self.polynomial.fit_transform(x)
        return super(ScikitLearner, self).fit(x=x, y=y, sample_weight=sample_weight)

    def predict(self, x) -> np.ndarray:
        x = x if self.polynomial is None else self.polynomial.transform(x)
        return super(ScikitLearner, self).predict(x=x)

    def _fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None):
        self.model.fit(x, y, sample_weight=sample_weight)

    def _predict(self, x) -> np.ndarray:
        return self.model.predict(x)


class ScikitClassifier(ScikitLearner):
    """Wrapper for a custom Scikit-Learn classification model."""

    def __init__(self,
                 model,
                 task: str = 'auto',
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **fit_kwargs):
        """
        :param model:
            The Scikit-Learn model.

        :param task:
            The kind of classification task, which can be either 'binary' (i.e., probabilities will be returned as a
            one-dimensional array), 'multiclass' (i.e., probabilities will be returned in a bi-dimensional array), or
            'auto' for automatic task detection.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param fit_kwargs:
            Custom arguments to be passed to the model '.fit()' method.
        """
        super(ScikitClassifier, self).__init__(model=model,
                                               polynomial=polynomial,
                                               mask=mask,
                                               x_scaler=x_scaler,
                                               y_scaler=y_scaler,
                                               stats=stats,
                                               **fit_kwargs)

        assert task in ['binary', 'multiclass', 'auto'], f"'task' should be either 'binary' or 'multiclass', got {task}"

        self.task: str = task
        """The kind of classification task."""

    def _predict(self, x) -> np.ndarray:
        probabilities = self.model.predict_proba(x)
        if self.task == 'multiclass' or probabilities.shape[1] > 2:
            # return probabilities matrix if the task is explicitly multiclass or if the number of classes is > 2
            return probabilities
        else:
            # return probabilities vector otherwise (i.e., explicitly binary task or automatic with classes == 2)
            return probabilities[:, 1].squeeze()


class LinearRegression(ScikitLearner):
    """Scikit-Learn Linear Regression wrapper."""

    def __init__(self,
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param model_kwargs:
            Custom arguments to be passed to a sklearn.linear_model.LogisticRegression instance.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.
        """
        super(LinearRegression, self).__init__(model=lm.LinearRegression(**model_kwargs),
                                               polynomial=polynomial,
                                               mask=mask,
                                               x_scaler=x_scaler,
                                               y_scaler=y_scaler,
                                               stats=stats)


class LogisticRegression(ScikitClassifier):
    """Scikit-Learn Logistic Regression wrapper."""

    def __init__(self,
                 task: str = 'auto',
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param task:
            The kind of classification task, which can be either 'binary' (i.e., probabilities will be returned as a
            one-dimensional array), 'multiclass' (i.e., probabilities will be returned in a bi-dimensional array), or
            'auto' for automatic task detection.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param model_kwargs:
            Custom arguments to be passed to a sklearn.linear_model.LogisticRegression instance.
        """
        super(LogisticRegression, self).__init__(model=lm.LogisticRegression(**model_kwargs),
                                                 task=task,
                                                 polynomial=polynomial,
                                                 mask=mask,
                                                 x_scaler=x_scaler,
                                                 y_scaler=y_scaler,
                                                 stats=stats)


class RandomForestRegressor(ScikitLearner):
    """Scikit-Learn Random Forest Regressor wrapper."""

    def __init__(self,
                 n_estimators: int = 100,
                 max_depth: Optional[int] = None,
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param n_estimators:
            The number of trees in the forest.

        :param max_depth:
            The maximum depth of the tree. If None, then nodes are expanded until all leaves are pure or until all
            leaves contain less than min_samples_split samples.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param model_kwargs:
            Additional arguments to be passed to a sklearn.ensemble.RandomForestRegressor instance.
        """
        m = ens.RandomForestRegressor(n_estimators=n_estimators, max_depth=max_depth, **model_kwargs)
        super(RandomForestRegressor, self).__init__(model=m,
                                                    polynomial=polynomial,
                                                    mask=mask,
                                                    x_scaler=x_scaler,
                                                    y_scaler=y_scaler,
                                                    stats=stats)


class RandomForestClassifier(ScikitClassifier):
    """Scikit-Learn Random Forest Classifier wrapper."""

    def __init__(self,
                 task: str = 'auto',
                 n_estimators: int = 100,
                 max_depth: Optional[int] = None,
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param task:
            The kind of classification task, which can be either 'binary' (i.e., probabilities will be returned as a
            one-dimensional array), 'multiclass' (i.e., probabilities will be returned in a bi-dimensional array), or
            'auto' for automatic task detection.

        :param n_estimators:
            The number of trees in the forest.

        :param max_depth:
            The maximum depth of the tree. If None, then nodes are expanded until all leaves are pure or until all
            leaves contain less than min_samples_split samples.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param model_kwargs:
            Additional arguments to be passed to a sklearn.ensemble.RandomForestClassifier instance.
        """
        m = ens.RandomForestClassifier(n_estimators=n_estimators, max_depth=max_depth, **model_kwargs)
        super(RandomForestClassifier, self).__init__(model=m,
                                                     task=task,
                                                     polynomial=polynomial,
                                                     mask=mask,
                                                     x_scaler=x_scaler,
                                                     y_scaler=y_scaler,
                                                     stats=stats)


class GradientBoostingRegressor(ScikitLearner):
    """Scikit-Learn Gradient Boosting Regressor wrapper."""

    def __init__(self,
                 n_estimators: int = 100,
                 min_samples_leaf: Union[int, float] = 1,
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param n_estimators:
            The number of boosting stages to perform.

        :param min_samples_leaf:
            The minimum number of samples required to be at a leaf node. If int, then consider `min_samples_leaf` as
            the minimum number, otherwise `min_samples_leaf` is a fraction and `ceil(min_samples_leaf * n_samples)` are
            the minimum number of samples for each node.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param model_kwargs:
            Additional arguments to be passed to a sklearn.ensemble.GradientBoostingRegressor instance.
        """
        m = ens.GradientBoostingRegressor(n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, **model_kwargs)
        super(GradientBoostingRegressor, self).__init__(model=m,
                                                        polynomial=polynomial,
                                                        mask=mask,
                                                        x_scaler=x_scaler,
                                                        y_scaler=y_scaler,
                                                        stats=stats)


class GradientBoostingClassifier(ScikitClassifier):
    """Scikit-Learn Gradient Boosting Classifier wrapper."""

    def __init__(self,
                 task: str = 'auto',
                 n_estimators: int = 100,
                 min_samples_leaf: Union[int, float] = 1,
                 polynomial: Optional[int] = None,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **model_kwargs):
        """
        :param task:
            The kind of classification task, which can be either 'binary' (i.e., probabilities will be returned as a
            one-dimensional array), 'multiclass' (i.e., probabilities will be returned in a bi-dimensional array), or
            'auto' for automatic task detection.

        :param n_estimators:
            The number of boosting stages to perform.

        :param min_samples_leaf:
            The minimum number of samples required to be at a leaf node. If int, then consider `min_samples_leaf` as
            the minimum number, otherwise `min_samples_leaf` is a fraction and `ceil(min_samples_leaf * n_samples)` are
            the minimum number of samples for each node.

        :param polynomial:
            The (optional) degree used to generate polynomial features. If not None, creates a scikit learn `Pipeline`
            object containing a `PolynomialFeatures` preprocessor (with given degree and no bias) at the beginning,
            followed by the given model.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param model_kwargs:
            Additional arguments to be passed to a sklearn.ensemble.GradientBoostingClassifier instance.
        """
        m = ens.GradientBoostingClassifier(n_estimators=n_estimators, min_samples_leaf=min_samples_leaf, **model_kwargs)
        super(GradientBoostingClassifier, self).__init__(model=m,
                                                         task=task,
                                                         polynomial=polynomial,
                                                         mask=mask,
                                                         x_scaler=x_scaler,
                                                         y_scaler=y_scaler,
                                                         stats=stats)



"""Classification Metrics."""

from typing import Callable, Optional, Dict, Any

import numpy as np

from moving_targets.metrics.metric import Metric
from moving_targets.util import probabilities


class ClassificationMetric(Metric):
    """Basic interface for a Moving Targets Classification Metric."""

    def __init__(self,
                 metric_function: Callable,
                 classes: Optional[int],
                 name: str,
                 use_prob: bool = False,
                 **metric_kwargs):
        """
        :param metric_function:
            Callable function that computes the metric given the true targets and the predictions.

        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param use_prob:
            Whether to use the output labels or the output probabilities.

        :param metric_kwargs:
            Custom kwargs to be passed to the metric function.
        """
        super(ClassificationMetric, self).__init__(name=name)

        self.metric_function: Callable = metric_function
        """Callable function that computes the metric given the true targets and the predictions."""

        self.classes: Optional[int] = classes
        """The number of classes or None for automatic class inference."""

        self.use_prob: bool = use_prob
        """Whether to use the output labels or the output probabilities."""

        self.metric_kwargs: Dict[str, Any] = metric_kwargs
        """Custom arguments to be passed to the metric function."""

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> float:
        if not self.use_prob:
            # if the metric does not use probabilities, convert the probabilities into class targets
            p = probabilities.get_classes(prob=p)
        elif p.ndim == 1 and np.issubdtype(p.dtype, np.integer):
            # otherwise, if we use class probabilities we expect a bi-dimensional array for predictions, thus if a
            # one-dimensional array is passed instead because the master directly returns adjusted class targets,
            # we onehot encode them for compatibility
            c = self.classes or probabilities.count_classes(vector=y)
            p = probabilities.get_onehot(vector=p, classes=c)
        return self.metric_function(y, p, **self.metric_kwargs)


class CrossEntropy(ClassificationMetric):
    """Wrapper for scikit-learn 'log_loss' function."""

    def __init__(self, classes: Optional[int] = None, name: str = 'crossentropy', **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'log_loss' function.
        """
        from sklearn.metrics import log_loss
        super(CrossEntropy, self).__init__(
            metric_function=log_loss,
            classes=classes,
            name=name,
            use_prob=True,
            **scikit_kwargs
        )


class Precision(ClassificationMetric):
    """Wrapper for scikit-learn 'precision_score' function."""

    def __init__(self, classes: Optional[int] = None, name: str = 'precision', **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'precision_score' function.
        """
        from sklearn.metrics import precision_score
        super(Precision, self).__init__(
            metric_function=precision_score,
            classes=classes,
            name=name,
            use_prob=False,
            **scikit_kwargs
        )


class Recall(ClassificationMetric):
    """Wrapper for scikit-learn 'recall_score' function."""

    def __init__(self, classes: Optional[int] = None, name: str = 'recall', **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'recall_score' function.
        """
        from sklearn.metrics import recall_score
        super(Recall, self).__init__(
            metric_function=recall_score,
            classes=classes,
            name=name,
            use_prob=False,
            **scikit_kwargs
        )


class F1(ClassificationMetric):
    """Wrapper for scikit-learn 'f1_score' function."""

    def __init__(self, classes: Optional[int] = None, name: str = 'f1', **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'f1_score' function.
        """
        from sklearn.metrics import f1_score
        super(F1, self).__init__(
            metric_function=f1_score,
            classes=classes,
            name=name,
            use_prob=False,
            **scikit_kwargs
        )


class Accuracy(ClassificationMetric):
    """Wrapper for scikit-learn 'accuracy_score' function."""

    def __init__(self, classes: Optional[int] = None, name: str = 'accuracy', **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Custom arguments to be passed to scikit-learn 'accuracy_score' function.
        """
        from sklearn.metrics import accuracy_score
        super(Accuracy, self).__init__(
            metric_function=accuracy_score,
            classes=classes,
            name=name,
            use_prob=False,
            **scikit_kwargs
        )


class AUC(ClassificationMetric):
    """Wrapper for scikit-learn 'roc_auc_score' function."""

    def __init__(self,
                 classes: Optional[int] = None,
                 multi_class: str = 'ovo',
                 name: str = 'auc',
                 **scikit_kwargs):
        """
        :param classes:
            The number of classes or None for automatic class inference.

        :param multi_class:
            Determines the type of configuration to use when dealing with multiclass targets, either 'ovr' or 'ovo'.

        :param name:
            The name of the metric.

        :param scikit_kwargs:
            Additional custom arguments to be passed to scikit-learn 'roc_auc_score' function.
        """
        from sklearn.metrics import roc_auc_score
        super(AUC, self).__init__(
            metric_function=roc_auc_score,
            classes=classes,
            name=name,
            use_prob=True,
            multi_class=multi_class,
            **scikit_kwargs
        )



"""Core of the Moving Targets algorithm."""
import time
import warnings
from typing import List, Dict, Callable, Union, Optional, Any, Set

import numpy as np

from moving_targets.callbacks.callback import Callback
from moving_targets.callbacks.console_logger import ConsoleLogger
from moving_targets.callbacks.file_logger import FileLogger
from moving_targets.callbacks.history import History
from moving_targets.callbacks.logger import Logger, StatsLogger
from moving_targets.learners.learner import Learner
from moving_targets.masters.masters import Master
from moving_targets.metrics.metric import Metric
from moving_targets.util.masking import mask_data, get_mask
from moving_targets.util.typing import Dataset, is_numeric


class MACS(StatsLogger):
    """Model-Agnostic Constraint Satisfaction.

    This class contains the core algorithm of Moving Targets. It leverages the `Learner` and `Master` instances to
    iteratively refine the predictions during the training phase and, eventually, it evaluates the solutions based on
    the given list of `Metric` objects.
    """

    @staticmethod
    def _parameters() -> Set[str]:
        return {'iteration', 'elapsed_time'}

    def __init__(self,
                 master: Master,
                 learner: Learner,
                 init_step: str = 'pretraining',
                 metrics: List[Metric] = (),
                 mask: Union[None, float, str] = 'learner',
                 stats: Union[bool, List[str]] = False):
        """
        :param master:
            A `Master` instance.

        :param learner:
            A `Learner` instance.

        :param init_step:
            The initial step of the algorithm, which can be either 'pretraining' or 'projection'.

        :param metrics:
            A list of `Metric` instances to evaluate the final solution.

        :param mask:
            Either an (optional) masking value, or a string between 'learner' and 'master' to use the same mask of the
            respective object. The MACS masking value will not be used during the actual fitting process, but rather
            it will be passed to the metric objects.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters in
            ['iteration', 'elapsed_time'] whose statistics must be logged.
        """
        super(MACS, self).__init__(stats=stats, name='MACS')
        assert init_step in ['pretraining', 'projection'], f"'{init_step}' is not a valid initial step"

        # handle mask
        if mask == 'learner':
            mask = learner.mask
        elif mask == 'master':
            mask = master.mask
        elif not (mask is None or is_numeric(mask)):
            raise AssertionError(f'Unexpected mask "{mask}", possible values are None, "learner", "master", or a float')

        self.master: Master = master
        """The `Master` instance."""

        self.learner: Learner = learner
        """The `Learner` instance."""

        self.init_step: str = init_step
        """The initial step of the algorithm, which can be either 'pretraining' or 'projection'."""

        self.iteration: Optional[int] = None
        """The current `MACS` iteration."""

        self.fitted: bool = False
        """Whether or not the `Learner` has been fitted at least once."""

        self.metrics: List[Metric] = list(metrics)
        """A list of `Metric` instances to evaluate the final solution."""

        self.mask: Optional[float] = mask
        """The (optional) masking value used to mask the original targets."""

        self._history: History = History()
        """The internal `History` object which is returned at the end of the training."""

        self._time: Optional[float] = None
        """An auxiliary variable to keep track of the elapsed time between iterations."""

    def fit(self,
            x,
            y: np.ndarray,
            iterations: int = 1,
            sample_weight: Optional[np.ndarray] = None,
            val_data: Optional[Dataset] = None,
            callbacks: List[Callback] = (),
            verbose: Union[int, bool] = 2) -> History:
        """Fits the `Learner` based on the Moving Targets iterative procedure.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param iterations:
            The number of algorithm iterations.

        :param sample_weight:
            The (optional) array of sample weights.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

            .. code-block:: python

                val_data = dict(train=(xtr, ytr), validation=(xvl, yvl), test=(xts, yts))

        :param callbacks:
            A list of `Callback` instances.

        :param verbose:
            Either a boolean or an int representing the verbosity value, such that:

            - `0` or `False` create no logger;
            - `1` creates a simple console logger with elapsed time only;
            - `2` or `True` create a more complete console logger with cached data at the end of each iterations.

        :return:
            An instance of the `History` object containing the training history.

        :raise `AssertionError`:
            If the number of iteration is negative, or is zero and the initial step is 'pretraining'.
        """
        # check user input
        assert iterations >= 0, f'the number of iterations should be non-negative, but it is {iterations}'
        assert self.init_step == 'pretraining' or iterations > 0, 'if projection, iterations cannot be zero'
        assert verbose in [True, False, 0, 1, 2], 'unknown verbosity'
        val_data = {} if val_data is None else (val_data if isinstance(val_data, dict) else {'val': val_data})

        # handle callbacks and verbosity (check for type instance since 1 == True and vice versa)
        callbacks = list(callbacks)
        if verbose == 1 and not isinstance(verbose, bool):
            callbacks += [ConsoleLogger()]
        elif verbose == 2 or verbose is True:
            callbacks += [FileLogger(routines=['on_pretraining_end', 'on_iteration_end'])]
        self._update(callbacks, lambda c: c.on_process_start(macs=self, x=x, y=y, val_data=val_data))

        # handle pretraining
        p = None
        if self.init_step == 'pretraining':
            self.iteration = 0
            self._update(callbacks, lambda c: c.on_pretraining_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self.learner.fit(x=x, y=y, sample_weight=sample_weight)
            self.fitted = True
            p = self.learner.predict(x)
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self._update(callbacks, lambda c: c.on_pretraining_end(macs=self, x=x, y=y, p=p, val_data=val_data))

        # algorithm core
        for self.iteration in range(1, iterations + 1):
            self._update(callbacks, lambda c: c.on_iteration_start(macs=self, x=x, y=y, val_data=val_data))
            self._update(callbacks, lambda c: c.on_adjustment_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- MASTER  STEP ----------------------------------------------
            z = self.master.adjust_targets(x=x, y=y, p=p, sample_weight=sample_weight)
            if z is None:
                # in case of no valid solution, raise a warning and stop the training loop
                warnings.warn(f'No solution found at iteration {self.iteration}, stop training. ' +
                              f'This may have been cause either by time limit or by model infeasibility.')
                break
            # ---------------------------------------------- MASTER  STEP ----------------------------------------------
            self._update(callbacks, lambda c: c.on_adjustment_end(macs=self, x=x, y=y, z=z, val_data=val_data))
            self._update(callbacks, lambda c: c.on_training_start(macs=self, x=x, y=y, val_data=val_data))
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self.learner.fit(x=x, y=z, sample_weight=sample_weight)
            self.fitted = True
            p = self.learner.predict(x)
            # ---------------------------------------------- LEARNER STEP ----------------------------------------------
            self._update(callbacks, lambda c: c.on_training_end(macs=self, x=x, y=y, p=p, val_data=val_data))
            self._update(callbacks, lambda c: c.on_iteration_end(macs=self, x=x, y=y, val_data=val_data))
        self._update(callbacks, lambda c: c.on_process_end(macs=self, x=x, y=y, val_data=val_data))
        return self._history

    def predict(self, x) -> Any:
        """Uses the previously trained `Learner` to predict labels from input samples.

        :param x:
            The matrix/dataframe of input samples.

        :return:
            The vector of predicted labels.

        :raise `AssertionError`:
            If the `Learner` has not been fitted yet.
        """
        assert self.fitted, 'The model has not been fitted yet, please call method .fit()'
        return self.learner.predict(x)

    def evaluate(self, x, y) -> Dict[str, float]:
        """Evaluates the performances of the model based on the given set of metrics.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :return:
            The dictionary of evaluated metrics.

        :raise `AssertionError`:
            If the `Learner` has not been fitted yet.
        """
        return self._compute_metrics(x=x, y=y, p=self.predict(x), metrics=self.metrics, prefix=None)

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._log_stats(iteration=self.iteration)
        self._time = time.time()

    def on_adjustment_end(self, macs, x, y: np.ndarray, z: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on adjusted data
        self.log(**self._compute_metrics(x=x, y=y, p=z, metrics=self.metrics, prefix='adjusted'))

    def on_training_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on training data
        self.log(**self._compute_metrics(x=x, y=y, p=p, metrics=self.metrics, prefix='predictions'))

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # log metrics on validation data
        for name, (xv, yv) in val_data.items():
            self.log(**self._compute_metrics(x=xv, y=yv, p=self.predict(xv), metrics=self.metrics, prefix=name))
        # log stats
        self._log_stats(elapsed_time=time.time() - self._time)

    def _compute_metrics(self,
                         x,
                         y: np.ndarray,
                         p: np.ndarray,
                         metrics: List[Metric],
                         prefix: Optional[str] = None) -> Dict[str, float]:
        """Compute the metrics results on the given data.

        :param x:
            The input data.

        :param y:
            The output targets.

        :param p:
            The predicted targets.

        :param metrics:
            The list of `Metric` instances for the evaluation.

        :param prefix:
            A prefix to be prepended to the metric name in the results dictionary.

        :return:
            A dictionary of metric results indexed by metric name (optionally with a prefix).
        """
        # mask data based on the output target
        x, y, p = mask_data(x, y, p, mask=get_mask(y, self.mask))
        prefix = "" if prefix is None else f'{prefix}/'
        results = {}
        for metric in metrics:
            value = metric(x, y, p)
            if is_numeric(value):
                # if the metric is a single metric, directly assign the value
                results[f'{prefix}{metric.__name__}'] = value
            elif isinstance(value, dict):
                # if the metric is a multi metric, assign each value per feature by using an underscore as a separator
                for k, v in value.items():
                    results[f'{prefix}{metric.__name__}_{k}'] = v
            else:
                # handle array-like return type by giving a warning
                self.logger.warning(f"Metric '{metric.__name__}' should return either a float or a dictionary.")
                for i, v in enumerate(value):
                    results[f'{prefix}{metric.__name__}_{i}'] = v
        return results

    def _update(self, callbacks: List[Callback], routine: Callable):
        """Runs the given routine for each one of the given callbacks, plus the routine for the `MACS` object itself
        (which is run at the beginning) and the inner `History` callback (which is run at the end).

        :param callbacks:
            The list of callbacks specified during the `fit()` call.

        :param routine:
            The callback routine (e.g., <callback>.on_iteration_start()).
        """
        # run the callback routine for the MACS object itself plus the master and learner instances, for which we do
        # not need to call the 'log' method even though they are 'Logger' instances since their cache is already shared
        routine(self)
        routine(self.master)
        routine(self.learner)
        # run the callback routine for each external callback plus the history logger
        for callback in callbacks + [self._history]:
            # if the callback is a logger, share the internal cache via the 'log' method before calling the routine
            if isinstance(callback, Logger):
                callback.log(**self._cache)
            routine(callback)
        # eventually clear the internal cache
        self._cache = {}



"""File Logger Callback"""

import sys
from typing import List, Set, Optional

import numpy as np

from moving_targets.callbacks.logger import Logger
from moving_targets.util.typing import Dataset

SEPARATOR: str = '--------------------------------------------------'


class FileLogger(Logger):
    """Logs the training data and information on a specified filepath or on the standard output."""

    def __init__(self, filepath: Optional[str] = None, routines: Optional[List[str]] = None, log_empty: bool = False,
                 sort_keys: bool = False, separator: str = SEPARATOR, end: str = '\n'):
        """
        :param filepath:
            Path string in which to put the log. If None, writes in the standard output.

        :param routines:
            List of routine names after which log the cached data. If None, logs after each routine.

        :param log_empty:
            Whether or not to log the routine name if there is no cached data.

        :param sort_keys:
            Whether or not to sort the cache data by key before logging.

        :param separator:
            String separator written between each routine.

        :param end:
            Line end.
        """
        super(FileLogger, self).__init__()

        self.filepath: str = filepath
        """Path string in which to put the log. If None, writes in the standard output."""

        self.routines: Optional[Set[str]] = None if routines is None else set(routines)
        """List of routine names after which log the cached data. If None, logs after each routine."""

        self.log_empty: bool = log_empty
        """Whether or not to log the routine name if there is no cached data."""

        self.sort_keys: bool = sort_keys
        """Whether or not to sort the cache data by key before logging."""

        self.separator: str = separator
        """String separator written between each routine."""

        self.end: str = end
        """Line end."""

        self._logged_once: bool = False
        """An internal variable used to write the initial line separator."""

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        # reset file content for overwriting
        if self.filepath is not None:
            open(self.filepath, 'w').close()

        self._write_on_file(macs, 'PROCESS START', 'on_process_start')

    def on_process_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'PROCESS END', 'on_process_end')

    def on_pretraining_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'PRETRAINING START', 'on_pretraining_start')

    def on_pretraining_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'PRETRAINING END', 'on_pretraining_end')

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, f'ITERATION START', 'on_iteration_start')

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, f'ITERATION END', 'on_iteration_end')

    def on_training_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'TRAINING START', 'on_training_start')

    def on_training_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'TRAINING END', 'on_training_end')

    def on_adjustment_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'ADJUSTMENT START', 'on_adjustment_start')

    def on_adjustment_end(self, macs, x, y: np.ndarray, z: np.ndarray, val_data: Optional[Dataset]):
        self._write_on_file(macs, 'ADJUSTMENT END', 'on_adjustment_end')

    def _write_on_file(self, macs, message: str, routine_name: str):
        """This is the core function, it is in charge of opening the chosen file and write the cached data when needed.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param message:
            An initial message to be printed before the cached data.

        :param routine_name:
            The name of the routine that just ended.
        """
        # in case no routine is inserted (i.e., all of them are considered) and either we are asked to log
        # independently from the presence of cached data or there is actually some cached data to log
        if (self.routines is None or routine_name in self.routines) and (self.log_empty or len(self._cache) > 0):
            # open file
            file = sys.stdout if self.filepath is None else open(self.filepath, 'a', encoding='utf8')
            # write initial separator if needed
            if not self._logged_once:
                file.write(f'{self.separator}{self.end}')
                self._logged_once = True
            # write message and cached items if present
            file.write(f'ITERATION {macs.iteration:02} - {message}{self.end}')
            cache = {k: self._cache[k] for k in sorted(self._cache)} if self.sort_keys else self._cache
            for k, v in cache.items():
                file.write(f'> {str(k)} = {str(v)}{self.end}')
            # write write separator and empty cache
            file.write(f'{self.separator}{self.end}')
            self._cache = {}
            # close file if not stdout
            if self.filepath is not None:
                file.close()



"""Basic Callback Interface."""

from typing import Optional

import numpy as np

from moving_targets.util.typing import Dataset


class Callback:
    """Basic interface for a Moving Targets callback."""

    def __init__(self):
        """"""
        super(Callback, self).__init__()

    def on_process_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the end of the `MACS` fitting process.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).
        """
        pass

    def on_process_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the beginning of the `MACS` fitting process.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).
        """
        pass

    def on_pretraining_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the beginning of the `MACS` pretraining phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        self.on_iteration_start(macs=macs, x=x, y=y, val_data=val_data)
        self.on_training_start(macs=macs, x=x, y=y, val_data=val_data)

    def on_pretraining_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the end of the `MACS` pretraining phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param p:
            The vector of learners predictions.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        self.on_training_end(macs=macs, x=x, y=y, p=p, val_data=val_data)
        self.on_iteration_end(macs=macs, x=x, y=y, val_data=val_data)

    def on_iteration_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the beginning of a `MACS` iteration.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass

    def on_iteration_end(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the end of a `MACS` iteration.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass

    def on_training_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the beginning of a `MACS` training phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass

    def on_training_end(self, macs, x, y: np.ndarray, p: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the end of a `MACS` training phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param p:
            The vector of learners predictions.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass

    def on_adjustment_start(self, macs, x, y: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the beginning of a `MACS` adjustment phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass

    def on_adjustment_end(self, macs, x, y: np.ndarray, z: np.ndarray, val_data: Optional[Dataset]):
        """Routine called at the end of a `MACS` adjustment phase.

        :param macs:
            Reference to the `MACS` object encapsulating the master.

        :param x:
            The matrix/dataframe of training samples.

        :param y:
            The vector of training labels.

        :param z:
            The vector of adjusted labels.

        :param val_data:
            A dictionary containing the validation data, indicated as a tuple (xv, yv).

        """
        pass



from typing import Optional, List, Union, Dict, Any

import numpy as np

from moving_targets.learners import Learner
from moving_targets.util.errors import MissingDependencyError
from moving_targets.util.scalers import Scaler


class TensorflowLearner(Learner):
    """Wrapper for a custom Tensorflow/Keras model."""

    def __init__(self,
                 model,
                 loss: Any,
                 optimizer: Any = 'adam',
                 metrics: Optional[List] = None,
                 loss_weights: Union[None, List, Dict] = None,
                 weighted_metrics: Optional[List] = None,
                 run_eagerly: bool = False,
                 warm_start: Union[bool, int] = False,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **fit_kwargs):
        """
        :param model:
            The tensorflow/keras model structure.

        :param loss:
            The neural network loss function (either a string or a loss object).

        :param optimizer:
            The neural network optimizer (either a string or an optimizer object).

        :param metrics:
            The list of tensorflow/keras metrics for the evaluation phase.

        :param loss_weights:
            Optional list or dictionary specifying scalar coefficients to weight the loss contributions.

        :param weighted_metrics:
            List of metrics to be evaluated and weighted by `sample_weight` or `class_weight`.

        :param run_eagerly:
            Whether or not to run tensorflow in eager mode.

        :param warm_start:
            Handles the warming start policy after each iteration. If 0 (or False), the weights are reinitialized and
            the model is trained from scratch. If 1 (or True), only the optimized is reinitialized, while the weights
            are kept from the previous iteration. If 2, neither the weights nor the optimizer are reinitialized.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param fit_kwargs:
            Custom arguments to be passed to the model '.fit()' method.
        """
        super(TensorflowLearner, self).__init__(mask=mask, x_scaler=x_scaler, y_scaler=y_scaler, stats=stats)

        self.model = model
        """The tensorflow/keras model structure."""

        assert warm_start in [0, 1, 2, True, False], "'warm_start' must be either a boolean or an integer in {0, 1, 2}"
        self.warm_start: int = warm_start if isinstance(warm_start, int) else int(warm_start)
        """The warm start level."""

        self.compile_kwargs: Dict[str, Any] = dict(
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            loss_weights=loss_weights,
            weighted_metrics=weighted_metrics,
            run_eagerly=run_eagerly
        )
        """Custom arguments to be passed to the '.compile()' method."""

        self.fit_kwargs: Dict[str, Any] = fit_kwargs
        """Custom arguments to be passed to the model '.fit()' method."""

        # if the warm start involves the optimizer as well, pre-compile the model
        if self.warm_start == 2:
            self.model.compile(**self.compile_kwargs)

    def _fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None):
        if self.warm_start == 0:
            # leverage the 'clone_model' utility to create a copy of the model structure with uninitialized weights
            from tensorflow.python.keras.models import clone_model
            self.model = clone_model(self.model)
            self.model.compile(**self.compile_kwargs)
        elif self.warm_start == 1:
            self.model.compile(**self.compile_kwargs)
        self.model.fit(x, y, sample_weight=sample_weight, **self.fit_kwargs)

    def _predict(self, x) -> np.ndarray:
        return self.model.predict(x).squeeze()


class TensorflowMLP(TensorflowLearner):
    """Tensorflow/Keras Dense Neural Network Wrapper"""

    def __init__(self,
                 loss: Any,
                 output_units: int = 1,
                 output_activation: Optional[str] = None,
                 hidden_units: List[int] = (128,),
                 hidden_activation: Optional[str] = 'relu',
                 optimizer: Any = 'adam',
                 metrics: Optional[List] = None,
                 loss_weights: Union[None, List, Dict] = None,
                 weighted_metrics: Optional[List] = None,
                 run_eagerly: bool = False,
                 epochs: int = 1,
                 shuffle: bool = True,
                 validation_split: float = 0.,
                 batch_size: Optional[int] = None,
                 class_weight: Optional[Dict] = None,
                 callbacks: Optional[List] = None,
                 verbose: Union[bool, str] = 'auto',
                 warm_start: Union[bool, int] = False,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **layer_kwargs: Any):
        """
        :param loss:
            The neural network loss function (either a string or a loss object).

        :param output_units:
            The neural network number of output units.

        :param output_activation:
            The neural network output activation.

        :param hidden_units:
            The tuple of neural network hidden units.

        :param hidden_activation:
            The neural network hidden activations.

        :param optimizer:
            The neural network optimizer (either a string or an optimizer object).

        :param metrics:
            The list of tensorflow/keras metrics for the evaluation phase.

        :param loss_weights:
            Optional list or dictionary specifying scalar coefficients to weight the loss contributions.

        :param weighted_metrics:
            List of metrics to be evaluated and weighted by `sample_weight` or `class_weight`.

        :param run_eagerly:
            Whether or not to run tensorflow in eager mode.

        :param epochs:
            The number of training epochs.

        :param shuffle:
            Whether or not to shuffle the dataset when training.

        :param validation_split:
            The validation split for neural network training.

        :param batch_size:
            The batch size for neural network training.

        :param class_weight:
            Optional dictionary mapping class indices to a weight, used for weighting the loss function during training.

        :param callbacks:
            The list of tensorflow/keras callbacks for the training phase.

        :param verbose:
            Whether or not to print information during the neural network training.

        :param warm_start:
            Handles the warming start policy after each iteration. If 0 (or False), the weights are reinitialized and
            the model is trained from scratch. If 1 (or True), only the optimized is reinitialized, while the weights
            are kept from the previous iteration. If 2, neither the weights nor the optimizer are reinitialized.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param layer_kwargs:
            Additional arguments to be passed to the layer instances, both hidden and output. Examples are: use_bias,
            kernel_initializer, kernel_regularizer, kernel_constraint, etc.
        """
        try:
            from tensorflow.python.keras.layers import Dense
            from tensorflow.python.keras.models import Sequential
        except ModuleNotFoundError:
            raise MissingDependencyError(package='tensorflow')

        network = Sequential()
        for units in hidden_units:
            network.add(layer=Dense(units=units, activation=hidden_activation, **layer_kwargs))
        network.add(layer=Dense(units=output_units, activation=output_activation, **layer_kwargs))

        super(TensorflowMLP, self).__init__(model=network,
                                            loss=loss,
                                            optimizer=optimizer,
                                            metrics=metrics,
                                            loss_weights=loss_weights,
                                            weighted_metrics=weighted_metrics,
                                            run_eagerly=run_eagerly,
                                            epochs=epochs,
                                            shuffle=shuffle,
                                            batch_size=batch_size,
                                            validation_split=validation_split,
                                            callbacks=callbacks,
                                            class_weight=class_weight,
                                            verbose=verbose,
                                            warm_start=warm_start,
                                            mask=mask,
                                            x_scaler=x_scaler,
                                            y_scaler=y_scaler,
                                            stats=stats)



from typing import Optional

import numpy as np
import pandas as pd


def get_mask(reference: np.ndarray, mask: Optional[float], all_columns: bool = True) -> np.ndarray:
    """Returns a masking vector

    :param reference:
        The reference vector on which to check the given mask.

    :param mask:
        Either the masking value used to compute the mask from the reference vector, or None for no masking (in this
        case, a vector of True is returned).

    :param all_columns:
        In case of a reference array with more than one dimension, the reference is reshaped into (len(reference), -1).
        If this parameter is set to true, and index gets masked if all the columns match the mask value, otherwise an
        index gets masked if at least one of the columns match the mask value.

    :return:
        The masking vector.
    """
    if mask is None:
        return np.ones(len(reference), dtype=bool)
    reference = np.atleast_2d(reference).reshape((len(reference), -1))
    mask = np.isnan(reference) if np.isnan(mask) else np.isclose(reference, mask)
    mask = np.all(mask, axis=1) if all_columns else np.any(mask, axis=1)
    return ~mask


def mask_data(*data, mask: np.ndarray) -> list:
    """Masks a set of data using the given mask vector.

    :param data:
        The data to mask.

    :param mask:
        The mask vector.

    :return:
        The masked data.
    """
    return [d.iloc[mask] if isinstance(d, (pd.DataFrame, pd.Series)) else d[mask] for d in data]



"""Moving Targets Optimizers."""
from typing import Optional, List

import numpy as np

from moving_targets.util.errors import not_implemented_message


class Optimizer:
    """Basic interface for a Moving Targets Master Optimizer."""

    def __init__(self, base: float = 1):
        """
        :param base:
            A floating point number representing the base (initial) value.
        """

        self.base: float = base
        """The base value of the optimizer."""

        self.value: Optional[float] = None
        """The last value computed from the optimizer."""

        self.history: List[float] = []
        """The history of past values yielded by the optimizer since the last initialization."""

    def __call__(self, x, y: np.ndarray, p: np.ndarray) -> float:
        """Runs the optimization strategy then returns the computed value.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :param p:
            The `Learner` predictions.

        :return:
            The newly computed value.
        """
        assert self.base is not None, "There is not base value for the optimizer, please call the 'initialize' method"
        self.value = self._next(x=x, y=y, p=p)
        self.history.append(self.value)
        return self.value

    def _next(self, x, y: np.ndarray, p: np.ndarray) -> float:
        """Implements the optimization strategy by returning the next computed value.

        :param x:
            The training samples.

        :param y:
            The training labels.

        :param p:
            The `Learner` predictions.

        :return:
            The computed value.
        """
        raise NotImplementedError(not_implemented_message(name='_next'))


class Constant(Optimizer):
    """Optimizer which returns a constant value."""

    def _next(self, x, y: np.ndarray, p: np.ndarray):
        return self.base


class Harmonic(Optimizer):
    """Optimizer which updates the values according to the harmonic series, i.e., by dividing the base value for the
    iteration number. E.g., starting from 10, the sequence will be: 10 -> 5 -> 3.33 -> 2.5 -> 2 -> ..."""

    def _next(self, x, y: np.ndarray, p: np.ndarray):
        return self.base / (len(self.history) + 1)


class Geometric(Optimizer):
    """Optimizer which updates the values according to the harmonic series, i.e., by dividing the last value for the
    given slope. E.g., starting from 10 and with a slope of 2, the sequence will be: 10 -> 5 -> 2.5 -> 1.25 -> ..."""

    def __init__(self, base: float = 1, slope: float = 2):
        """
        :param base:
            A floating point number representing the base (initial) value.

        :param slope:
            The base of the slope used to decrease the hyper-parameter value.
        """
        super(Geometric, self).__init__(base=base)

        self.slope: float = slope
        """The slope used to decrease the hyper-parameter value."""

    def _next(self, x, y: np.ndarray, p: np.ndarray) -> float:
        return self.base if self.value is None else self.value / self.slope


aliases: dict = {
    'constant': Constant,
    'harmonic': Harmonic,
    'geometric': Geometric,
}
"""Dictionary which associates to each optimizer alias the respective optimizer type."""



"""Callback and Logger classes for Moving Targets."""

from moving_targets.callbacks.callback import Callback
from moving_targets.callbacks.console_logger import ConsoleLogger
from moving_targets.callbacks.data_logger import DataLogger
from moving_targets.callbacks.file_logger import FileLogger
from moving_targets.callbacks.history import History
from moving_targets.callbacks.logger import Logger, StatsLogger
from moving_targets.callbacks.wandb_logger import WandBLogger



from typing import Optional, List, Union, Dict, Any, Tuple

import numpy as np

from moving_targets.learners import Learner
from moving_targets.util.errors import MissingDependencyError
from moving_targets.util.scalers import Scaler


class TorchLearner(Learner):
    """Wrapper for a custom Torch model."""

    class _Dataset:
        """Inner utility class to deal with torch expected input."""

        def __init__(self, x, y):
            import torch
            assert len(x) == len(y), f"Data should have the same length, but len(x) = {len(x)} and len(y) = {len(y)} "
            self.x = torch.tensor(x, dtype=torch.float32)
            self.y = torch.tensor(np.expand_dims(y, axis=-1), dtype=torch.float32)

        def __len__(self):
            return len(self.y)

        def __getitem__(self, idx):
            return self.x[idx], self.y[idx]

    def __init__(self,
                 model,
                 loss,
                 optimizer='Adam',
                 epochs: int = 1,
                 shuffle: bool = True,
                 validation_split: float = 0.,
                 batch_size: Optional[int] = 128,
                 num_workers: int = 0,
                 verbose: bool = True,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False,
                 **fit_kwargs):
        """
        :param model:
            The torch model which should have been already compiled.

        :param loss:
            Either the neural network loss function or its name to retrieve it from `torch.nn`.

        :param optimizer:
            Either the neural network optimizer or its name to retrieve it from `torch.optim`.

        :param epochs:
            The number of training epochs.

        :param shuffle:
            Whether or not to shuffle the dataset when training.

        :param validation_split:
            The validation split for neural network training.

        :param batch_size:
            The batch size for neural network training.

        :param num_workers:
            Defines how many subprocesses to use for data loading ('0' to load data in the main process).

        :param verbose:
            Whether or not to print information during the neural network training.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.

        :param fit_kwargs:
            Custom arguments to be passed to the model '.fit()' method.
        """
        try:
            from torch import nn, optim
        except ModuleNotFoundError:
            raise MissingDependencyError(package='torch')

        super(TorchLearner, self).__init__(mask=mask, x_scaler=x_scaler, y_scaler=y_scaler, stats=stats)

        # handle loss
        if isinstance(loss, str):
            loss = getattr(nn, loss)
            loss = loss()

        # handle optimizer
        if isinstance(optimizer, str):
            optimizer = getattr(optim, optimizer)
            optimizer = optimizer(model.parameters())

        self.model: nn.Module = model
        """The torch model."""

        self.loss: nn.Module = loss
        """The neural network loss function."""

        self.optimizer: optim.Optimizer = optimizer
        """The neural network optimizer."""

        self.epochs: int = epochs
        """The number of training epochs."""

        self.shuffle: bool = shuffle
        """Whether or not to shuffle the dataset when training."""

        self.validation_split: float = validation_split
        """The validation split for neural network training."""

        self.batch_size: int = batch_size
        """The batch size for neural network training."""

        self.num_workers: int = num_workers
        """The number of subprocesses to use for data loading."""

        self.verbose: bool = verbose
        """Whether or not to print information during the neural network training."""

        self.fit_kwargs: Dict[str, Any] = fit_kwargs
        """Custom arguments to be passed to the model '.fit()' method."""

    def _print(self, epoch: int, loss: float, batch: Optional[Tuple[int, int]] = None):
        if not self.verbose:
            return

        # carriage return to write in the same line
        print(f'\r', end='')
        # print the epoch number with trailing spaces on the left to match the maximum epoch
        print(f'Epoch {epoch:{len(str(self.epochs))}}', end=' ')
        # print the loss value (either loss for this single batch or for the whole epoch)
        print(f'- loss = {loss:.4f}', end='')
        # check whether this is the information of a single batch or of the whole epoch and for the latter case (batch
        # is None) print a new line, while for the former (batch is a tuple) print the batch number and no new line
        if batch is None:
            print()
        else:
            batch, batches = batch
            print(f' (batch {batch:{len(str(batches))}} of {batches})', end='')

    # noinspection PyTypeChecker
    def _fit(self, x, y: np.ndarray, sample_weight: Optional[np.ndarray] = None):
        from torch.utils.data import DataLoader

        if sample_weight is not None:
            self.logger.warning("TorchLearner does not support sample weights, please pass 'sample_weight'=None")

        ds = TorchLearner._Dataset(x=x, y=y)
        loader = DataLoader(dataset=ds, batch_size=self.batch_size, shuffle=self.shuffle, num_workers=self.num_workers)
        batches = np.ceil(len(ds) / self.batch_size).astype(int)
        self.model.train()
        for epoch in np.arange(self.epochs) + 1:
            epoch_loss = 0.0
            for batch, (inp, out) in enumerate(loader):
                self.optimizer.zero_grad()
                pred = self.model(inp)
                loss = self.loss(pred, out)
                loss.backward()
                self.optimizer.step()
                epoch_loss += loss.item() * inp.size(0)
                self._print(epoch=epoch, loss=loss.item(), batch=(batch + 1, batches))
            self._print(epoch=epoch, loss=epoch_loss / len(ds))
        self.model.eval()

    def _predict(self, x) -> np.ndarray:
        import torch
        return self.model(torch.tensor(x, dtype=torch.float32)).detach().numpy().squeeze()


class TorchMLP(TorchLearner):
    """Torch Dense Neural Network Wrapper"""

    def __init__(self,
                 loss,
                 input_units: int,
                 output_units: int = 1,
                 output_activation: Optional = None,
                 hidden_units: List[int] = (128,),
                 hidden_activation: Optional = 'ReLU',
                 optimizer: str = 'Adam',
                 epochs: int = 1,
                 shuffle: bool = True,
                 batch_size: Optional[int] = 128,
                 num_workers: int = 0,
                 verbose: bool = True,
                 mask: Optional[float] = None,
                 x_scaler: Union[None, Scaler, str] = None,
                 y_scaler: Union[None, Scaler, str] = None,
                 stats: Union[bool, List[str]] = False):
        """
        :param loss:
            Either the neural network loss function or its name to retrieve it from `torch.nn`.

        :param input_units:
            The neural network number of input units.

        :param output_units:
            The neural network number of output units.

        :param output_activation:
            Either the neural network output activation module or its name to retrieve it from `torch.nn`.

        :param hidden_units:
            The tuple of neural network hidden units.

        :param hidden_activation:
            Either the neural network hidden activation module or its name to retrieve it from `torch.nn`.

        :param optimizer:
            The name of the neural network optimizer to retrieve it from `torch.optim`.

        :param epochs:
            The number of training epochs.

        :param shuffle:
            Whether or not to shuffle the dataset when training.

        :param batch_size:
            The batch size for neural network training.

        :param num_workers:
            Defines how many subprocesses to use for data loading ('0' to load data in the main process).

        :param verbose:
            Whether or not to print information during the neural network training.

        :param mask:
            The (optional) masking value used to mask the original targets.

        :param x_scaler:
            The (optional) scaler for the input data, or a string representing the default scaling method.

        :param y_scaler:
            The (optional) scaler for the output data, or a string representing the default scaling method.

        :param stats:
            Either a boolean value indicating whether or not to log statistics, or a list of parameters whose
            statistics must be logged.
        """
        try:
            from torch import nn
        except ModuleNotFoundError:
            raise MissingDependencyError(package='torch')

        # handle hidden activation
        if isinstance(hidden_activation, str):
            hidden_activation = getattr(nn, hidden_activation)
            hidden_activation = hidden_activation()

        # handle output activation
        if isinstance(output_activation, str):
            output_activation = getattr(nn, output_activation)
            output_activation = output_activation()

        # build the hidden linear layers and optionally append an activation layer if passed
        layers = []
        hidden_units = [input_units] + list(hidden_units)
        for i, h in enumerate(hidden_units[:-1]):
            layers += [nn.Linear(in_features=hidden_units[i], out_features=hidden_units[i + 1])]
            layers += [] if hidden_activation is None else [hidden_activation]
        # build the output layer and optionally append an output activation layer if passed
        layers += [nn.Linear(in_features=hidden_units[-1], out_features=output_units)]
        layers += [] if output_activation is None else [output_activation]
        # eventually build the network
        network = nn.Sequential(*layers)

        super(TorchMLP, self).__init__(model=network,
                                       loss=loss,
                                       optimizer=optimizer,
                                       epochs=epochs,
                                       shuffle=shuffle,
                                       batch_size=batch_size,
                                       num_workers=num_workers,
                                       verbose=verbose,
                                       mask=mask,
                                       x_scaler=x_scaler,
                                       y_scaler=y_scaler,
                                       stats=stats)
